package fr.doap.jdb;

import fr.doap.jdb.utils.*;
import fr.doap.slog.*;

import java.lang.reflect.*;
import java.sql.*;
import java.util.*;
import java.util.concurrent.*;

public class JDB {
	/* public static */
	static public void init(String aConn, String aUser, String aPass) {
		sJDB = new JDB();
		sJDB._init(aConn, aUser, aPass);
	}
	
	static public void save(Object aObject) {
		sJDB._save(aObject);
	}
	
	static protected JDB sJDB;
	
	/* protected member */
	protected void _init(String aConn, String aUser, String aPass) {
		mDBWrap = new DBWrap(aConn, aUser, aPass).setSQLTypeResolver(aCol -> mSQLTypes.get(aCol));
		mSQLTypes.put("_CLASS_", "VARCHAR");
		mSQLTypes.put("_SUBS_", "VARBINARY");
		
		Optional<Map<String, Object>> lOpt = mDBWrap.query("SELECT MAX("+DBWrap.sJID+")+1 AS NEXTID FROM OBJECTS").findFirst();
		Object lNID = lOpt.get().get("NEXTID");
		if (lNID != null) mDBWOC.setNextID((int)lNID);
		else mDBWOC.setNextID(1);
	}
	
	protected void _save(Object aObject) {
		Objects2Maps lO2M = new Objects2Maps(mDBWOC);
		lO2M.add(aObject);
		mSQLTypes.putAll(lO2M.getColumnSQLTypes());
		lO2M.getInserts().values().forEach(aProps -> mDBWrap.insert(aProps));
		lO2M.getUpdates().values().forEach(aProps -> mDBWrap.update(aProps));
	}
	
	// aConds = String: le SQL et ensuite les paramètres
	protected <T> Iterator<T> _query(Class<T> aClass, Object ... aConds) {
		return new Iterator<T>() {
			ResultSet mRS;
			{
				try {
					mRS = _buildQuery(Utils.getShortClassname(aClass), aConds).get();
				}
				catch (Exception aE) { L.w(aE); }
			}
			
			@Override
			public boolean hasNext() {
				try { return !(mRS.isLast() || mRS.isAfterLast()); }
				catch (Exception aE) { return false; }
			}
			
			@Override
			public T next() {
				try {
					mRS.next();
					int lID = mRS.getInt(mDBWrap.sJID);
L.i("ID: {}", lID);
					Object lCached = mDBWOC.getObject(lID);
					if (lCached != null) return (T)lCached;
L.i("NOT CACHED ! Loading...");
					
					Map<Integer, Map<String, Object>> lProps = new HashMap<>();
					Set<String> lCols = new HashSet<>();
					int lColCount = mRS.getMetaData().getColumnCount();
					for (int i=0; i<lColCount; i++) lCols.add(mRS.getMetaData().getColumnName(i+1));
L.i("Cols: {}", lCols);
					
					Map<String, Object> lMap = new HashMap<>();
					lProps.put(lID, lMap);
					for (String iCol: lCols) {
						Object lP = mRS.getObject(iCol);
						if (lP != null) lMap.put(iCol, lP);
					}
L.i("Props: {}", lMap);
					
					List<Integer> lLoadList = new ArrayList<>();
					byte[] lBA = mRS.getBytes("_SUBS_");
					if (lBA != null) {
						int[] lIDs = (int[]) Utils.fromByteArray(lBA);
						for (int i=0; i<lIDs.length; i++) if (mDBWOC.getObject(lIDs[i]) == null) lLoadList.add(lIDs[i]);
					}
					
					ResultSet lLoadRS = mDBWrap._query("SELECT * FROM OBJECTS WHERE \""+DBWrap.sJID+"\" IN ("+String.join(",", lLoadList.stream().map(aID -> aID.toString()).toList())+")").get();
					while (!lLoadRS.isLast() && !lLoadRS.isAfterLast()) {
						lLoadRS.next();
						
						lMap = new HashMap<>();
						int lSID = lLoadRS.getInt(mDBWrap.sJID);
						lProps.put(lSID, lMap);
						for (String iCol: lCols) {
							Object lP = lLoadRS.getObject(iCol);
							if (lP != null) lMap.put(iCol, lP);
						}
L.i(" --> Props: {}", lMap);
						
						Class<?> lClass = Class.forName(lLoadRS.getString("_CLASS_"));
						Object lO = lClass.getDeclaredConstructor().newInstance();
						mDBWOC.put(lSID, lO);
					}
					lProps.forEach((aID, aProps) -> {
						Object lO = mDBWOC.getId(aID);
						if ((lO instanceof Collection)&&(lO.getClass().getCanonicalName().startsWith("java."))) {
						
						} else {
							aProps.forEach((aK, aV) -> {
								if (!aK.startsWith("_") && aK.contains("#")) {
									String[] lKs = aK.split("#");
									try {
										Field lF = Class.forName(lKs[0]).getDeclaredField(lKs[1]);
										L.i("- key: {} -> {}", aK, lF);
									} catch (Exception aE) { L.t(aE); }
								}
							});
						}
					});
					
					return (T) mDBWOC.getObject(lID);
				} catch (Exception aE) { L.w(aE); }
				return null;
			}
		};
	}
	
	protected Future<ResultSet> _buildQuery(String aClassname, Object ... aConds) {
		Object[] lParams = null;
		String lSQL = "SELECT * FROM OBJECTS WHERE \""+aClassname+"\"=TRUE";
		if ((aConds != null)&&(aConds.length > 0)) {
			lSQL += " AND "+(String)aConds[0];
			if (aConds.length > 1) {
				lParams = new Object[aConds.length - 1];
				for (int i = 0; i < lParams.length; i++) lParams[i] = aConds[i + 1];
			}
		}
		return mDBWrap._query(lSQL, lParams);
	}
	
	protected DBWrap				mDBWrap;
	protected DBWOC 				mDBWOC = new DBWOC();
	protected Map<String, String>	mSQLTypes = new HashMap<>();
}