package fr.doap.jdb.utils;

import java.util.*;
import java.util.function.*;

public class ObjectVisitor {
	
	protected boolean onCheckObjectBeforeAdding(Object aObject, int aPos) { return true; }
	protected void onObjectAdded(Object aObject, int aPos) {}
	
	protected void onStartingRun(int aPos) {}
	protected void onProcessingObject(Object aObject, int aPos) {}
	protected void onEndingRun(int aStart, int aPos) {}
	
	public void add(Consumer<Object> aCallback, Object ... aObjects) {
		boolean lNeedRun = false;
		if (aObjects != null) {
			for (Object iO : aObjects) {
				if (iO != null) {
					if (!mProcessingList.contains(iO)) {
						if (onCheckObjectBeforeAdding(iO, mProcessingList.size())) {
							lNeedRun = true;
							mProcessingList.add(iO);
							onObjectAdded(iO, mProcessingList.size()-1);
						}
					}
					Set<Consumer<Object>> lCallbacks = mPostProcessCallback.get(iO);
					if (lCallbacks == null) {
						lCallbacks = new HashSet<>();
						mPostProcessCallback.put(iO, lCallbacks);
					}
					lCallbacks.add(aCallback);
				}
			}
			if (lNeedRun) run();
		}
	}
	
	public void add(Object ... aObjects) {
		boolean lNeedRun = false;
		if (aObjects != null) {
			for (Object iO : aObjects) {
				if (iO != null) {
					if (!mProcessingList.contains(iO)) {
						if (onCheckObjectBeforeAdding(iO, mProcessingList.size())) {
							lNeedRun = true;
							mProcessingList.add(iO);
							onObjectAdded(iO, mProcessingList.size()-1);
						}
					}
				}
			}
			if (lNeedRun && !mRunning) run();
		}
	}
	
	protected void run() {
		mRunning = true;
		int lStartPos = mPos;
		onStartingRun(mPos);
		while (mPos < mProcessingList.size()) {
			Object lO = mProcessingList.get(mPos);
			onProcessingObject(lO, mPos);
			mPos++;
		}
		for (int iP=mPos-1; iP>=lStartPos; iP--) {
			Object lO = mProcessingList.get(iP);
			Set<Consumer<Object>> lCallbacks = mPostProcessCallback.get(lO);
			if (lCallbacks != null) lCallbacks.forEach(aCallback -> aCallback.accept(lO));
		}
		onEndingRun(lStartPos, mPos);
		mRunning = false;
	}

	protected boolean								mRunning = false;
	protected int									mPos = 0;
	protected List<Object>							mProcessingList = new ArrayList<>();
	protected Map<Object, Set<Consumer<Object>>>	mPostProcessCallback = new HashMap<>();
}
